"""
This file defines utils function specific to the library
"""

import numpy as np
from math import *
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button

from .utils import boundary_to_boundary_quasi_toep, sort, epsilon, eta_func, parametrization
from .boundaries import Dirichlet
from .complex_winding_number import *


def show():
    plt.show()


def detKLplotsimple(schem, n_param=300, parametrization_bool=True):
    """Computes the Kreiss-Lopatinskii determinant curve for a given scheme

    :param schem: Scheme considered
    :type schem: class:`Scheme`
    :param n_param: Size of the discretization of the unit circle, defaults to 300
    :type n_param: int, optional
    :param parametrization_bool: True to have a refinment close to the origin, False to have a uniform discretization of the unit circle, defaults to True
    :type parametrization_bool: bool, optional
    :return: the Kreiss-Lopatinskii curve
    :rtype: plot object
    """
    ax = plt.subplot()
    Param, Det = schem.detKL(n_param, parametrization_bool)
    ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2)
    ax.axvline(x=0, color="0.5")
    ax.axhline(y=0, color="0.5")
    ax.axis("equal")
    return ax


def detKLplot(schem, left_bound=Dirichlet(), lamb=None, sigma=0, lambdacursor=False, sigmacursor=False, nparam=300, parametrization_bool=True, order=2, fig_size=(6, 4)):
    """Computes the Kreiss-Lopatinskii determinant curve for different lambdas and different sigmas or with cursor(s)

    :param schem: Scheme depending on a parameter lambda
    :type schem: class:`Scheme`
    :param left_bound: Left boundary of the class Boundary, defaults to Dirichlet()
    :type left_bound: class:`Boundary`, optional
    :param lamb: A value lambda or a numpy.ndarray of several values of lambda (lambda is the Courant number a.dt/dx), defaults to None
    :type lamb: int or float or list or numpy.ndarray, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: int or float or list or numpy.ndarray, optional
    :param lambdacursor: Boolean to indicate the use of a cursor for lambda's moving among the lamb values, defaults to False
    :type lambdacursor: bool, optional
    :param sigmacursor: Boolean to indicate the use of a cursor for sigma's moving among the sigma values, defaults to False
    :type sigmacursor: bool, optional
    :param n_param: Size of the discretization of the unit circle, defaults to 300
    :type n_param: int, optional
    :param parametrization_bool: True to have a refinment close to the origin, False to have a uniform discretization of the unit circle, defaults to True
    :type parametrization_bool: bool, optional
    :param order: Order of the scheme (for Lax-Wendroff scheme), defaults to 2
    :type order: int, optional
    :param fig_size: Size of the figure, defaults to (6,4)
    :type fig_size: (int,int), optional
    :return: the Kreiss-Lopatinskii curve for the scheme
    :rtype: plot object
    """
    if lamb is None:
        lamb = np.array([schem(1).CFL])
    if isinstance(lamb, (int, float)):
        lamb = np.array([lamb], dtype=float)
    if isinstance(sigma, (int, float)):
        sigma = np.array([sigma], dtype=float)
    if sigmacursor == False:
        if lambdacursor == False:
            if len(sigma) == 1:
                lamb = np.sort(lamb)
                fig, ax = plt.subplots(figsize=fig_size)
                for l in lamb:
                    S = schem(l, left_bound, order=order, sigma=sigma[0])
                    Param, Det = S.detKL(nparam, parametrization_bool)
                    ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, label=f"$\lambda$ = " + str(l))
                ax.axvline(x=0, color="0.5")
                ax.axhline(y=0, color="0.5")
                ax.legend(loc="best")
                plt.title("Symbol of " + S.name(boundary_bool=True, sigma_bool=True, lambda_bool=False))
            elif len(lamb) == 1:
                sigma = np.sort(sigma)
                fig, ax = plt.subplots(figsize=fig_size)
                for sig in sigma:
                    S = schem(lamb[0], left_bound, order=order, sigma=sig)
                    Param, Det = S.detKL(nparam, parametrization_bool)
                    ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, label=f"$\sigma$ = " + str(sig))
                ax.axvline(x=0, color="0.5")
                ax.axhline(y=0, color="0.5")
                ax.legend(loc="best")
                plt.title("Symbol of " + S.name(boundary_bool=True, sigma_bool=False, lambda_bool=True))
            else:
                sigma = np.sort(sigma)
                fig, ax = plt.subplots(figsize=fig_size)
                for l in lamb:
                    for sig in sigma:
                        S = schem(l, left_bound, order=order, sigma=sig)
                        Param, Det = S.detKL(nparam, parametrization_bool)
                        ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, label=f"$\lambda$ = " + str(l) + f"et $\sigma$ = " + str(sig))
                ax.axvline(x=0, color="0.5")
                ax.axhline(y=0, color="0.5")
                ax.legend(loc="best")
                plt.title("Symbol of " + S.name(boundary_bool=True, sigma_bool=False, lambda_bool=False))
        else:
            if len(sigma) > 1:
                raise TypeError("with a lambda cursor, sigma has to be a single value")
            lambmax = np.max(lamb)
            if len(lamb) == 1 or np.min(lamb) < 0.001:
                lambmin = 0.001
            else:
                lambmin = np.min(lamb)

            lamb0 = (lambmax + lambmin) / 2
            fig = plt.figure(1, figsize=(10, 8))
            plt.title("DKL curve of " + schem(1, left_bound, sigma=sigma[0]).name(boundary_bool=True, sigma_bool=True))
            ax = fig.add_subplot(111)
            fig.subplots_adjust(left=0.25, bottom=0.25)
            ax.set_xlim(-3, 3)
            ax.axis("equal")

            def calc_det(l):
                S = schem(l, left_bound, sigma=sigma[0], order=order)
                return S.detKL(nparam, parametrization_bool)[1]

            Det = calc_det(lamb0)
            [line] = ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, color="red")
            ax.axvline(x=0, color="0.5")
            ax.axhline(y=0, color="0.5")
            ax.axis("equal")

            lambda_slider_ax = fig.add_axes([0.25, 0.15, 0.65, 0.03])
            lambda_slider = Slider(ax=lambda_slider_ax, label="lambda", valmin=lambmin, valmax=lambmax, valinit=lamb0)

            def update(val):
                Det = calc_det(lambda_slider.val)
                line.set_xdata([z.real for z in Det])
                line.set_ydata([z.imag for z in Det])
                xmin = min([z.real for z in Det])
                xmax = max([z.real for z in Det])
                ymin = min([z.imag for z in Det])
                ymax = max([z.imag for z in Det])
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                fig.canvas.draw_idle()

            # register the update function with each slider
            lambda_slider.on_changed(update)

            # Create a `matplotlib.widgets.Button` to reset the sliders to initial values.
            resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
            button = Button(resetax, "Reset", hovercolor="0.975")

            def reset(event):
                lambda_slider.reset()

            button.on_clicked(reset)

    else:
        if len(sigma) == 1:
            sigmax = 1 / 2
            sigmin = -1 / 2
        else:
            sigmax = np.max(sigma)
            sigmin = np.min(sigma)
        if lambdacursor == False:
            if len(lamb) > 1:
                raise TypeError("With a sigma cursor, lambda has to be a single value")
            sigma0 = (sigmax + sigmin) / 2
            fig = plt.figure(1, figsize=(10, 8))
            plt.title("DKL curve of " + schem(lamb[0], left_bound).name(boundary_bool=True, lambda_bool=True))
            ax = fig.add_subplot(111)
            fig.subplots_adjust(left=0.25, bottom=0.25)
            ax.set_xlim(-3, 3)
            ax.axis("equal")

            def calc_det(sig):
                S = schem(lamb[0], left_bound, sigma=sig, order=order)
                return S.detKL(nparam, parametrization_bool)[1]

            Det = calc_det(sigma0)
            [line] = ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, color="red")
            ax.axvline(x=0, color="0.5")
            ax.axhline(y=0, color="0.5")
            ax.axis("equal")

            sigma_slider_ax = fig.add_axes([0.25, 0.15, 0.65, 0.03])
            sigma_slider = Slider(ax=sigma_slider_ax, label="sigma", valmin=sigmin, valmax=sigmax, valinit=sigma0)

            def update(val):
                Det = calc_det(sigma_slider.val)
                line.set_xdata([z.real for z in Det])
                line.set_ydata([z.imag for z in Det])
                xmin = min([z.real for z in Det])
                xmax = max([z.real for z in Det])
                ymin = min([z.imag for z in Det])
                ymax = max([z.imag for z in Det])
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                fig.canvas.draw_idle()

            # register the update function with each slider
            sigma_slider.on_changed(update)

            # Create a `matplotlib.widgets.Button` to reset the sliders to initial values.
            resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
            button = Button(resetax, "Reset", hovercolor="0.975")

            def reset(event):
                sigma_slider.reset()

            button.on_clicked(reset)

        else:
            lambmax = np.max(lamb)
            if len(lamb) == 1 or np.min(lamb) < 0.001:
                lambmin = 0.001
            else:
                lambmin = np.min(lamb)

            sigma0 = (sigmax + sigmin) / 2
            lamb0 = (lambmin + lambmax) / 2

            fig = plt.figure(1, figsize=(10, 8))
            plt.title("DKL curve of " + schem(1, left_bound).name(boundary_bool=True))
            ax = fig.add_subplot(111)
            fig.subplots_adjust(left=0.25, bottom=0.25)
            ax.set_xlim(-3, 3)
            ax.axis("equal")

            def calc_det(l, sigm):
                S = schem(l, left_bound, sigma=sigm, order=order)
                return S.detKL(nparam, parametrization_bool)[1]

            Det = calc_det(lamb0, sigma0)
            [line] = ax.plot([z.real for z in Det], [z.imag for z in Det], linewidth=2, color="red")
            plt.axvline(x=0, color="0.5")
            plt.axhline(y=0, color="0.5")

            axlambda = plt.axes([0.25, 0.1, 0.65, 0.03])
            lambda_slider = Slider(
                ax=axlambda,
                label="lambda",
                valmin=lambmin,
                valmax=lambmax,
                valinit=lamb0,
            )

            # Make a vertically oriented slider to control the amplitude
            axsigm = plt.axes([0.1, 0.25, 0.0225, 0.63])
            sigma_slider = Slider(ax=axsigm, label="sigma", valmin=sigmin, valmax=sigmax, valinit=sigma0, orientation="vertical")

            def update(val):
                Det = calc_det(lambda_slider.val, sigma_slider.val)
                line.set_xdata([z.real for z in Det])
                line.set_ydata([z.imag for z in Det])
                xmin = min([z.real for z in Det])
                xmax = max([z.real for z in Det])
                ymin = min([z.imag for z in Det])
                ymax = max([z.imag for z in Det])
                ax.set_xlim(xmin, xmax)
                ax.set_ylim(ymin, ymax)
                fig.canvas.draw_idle()

            lambda_slider.on_changed(update)
            sigma_slider.on_changed(update)
            resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
            button = Button(resetax, "Reset", hovercolor="0.975")

            def reset(event):
                lambda_slider.reset()
                sigma_slider.reset()

            button.on_clicked(reset)

    return ax


def symbolplot(schem, lamb=None, order=2, lambdacursor=False, nparam=300, fig_size=(10, 8)):
    """Computes the Kreiss-Lopatinskii determinant curve for different lambdas and different sigmas or with cursor(s)

    :param schem: Scheme depending on a parameter lambda
    :type schem: class:`Scheme`
    :param lamb: A value lambda or a numpy.ndarray of several values of lambda (lambda is the Courant number a.dt/dx), defaults to None
    :type lamb: int or float or list or numpy.ndarray, optional
    :param order: Order of the scheme (for Lax-Wendroff scheme), defaults to 2
    :type order: int, optional
    :param lambdacursor: Boolean to indicate the use of a cursor for lambda's moving among the lamb values, defaults to False
    :type lambdacursor: bool, optional
    :param nparam: Size of the discretization of the unit circle, defaults to 300
    :type nparam: int, optional
    :param fig_size: Size of the figure, defaults to (10,8)
    :type fig_size: (int,int), optional
    :return: the symbol curve of the scheme
    :rtype: plot object
    """
    if lambdacursor == False:
        fig, ax = plt.subplots(figsize=fig_size)
        if isinstance(lamb, (int, float)):
            S = schem(lamb, order=order)
            X, Y = S.symbol(nparam)
            T = np.linspace(0, 7, 1000)

            plt.ylim(-1, 1)
            plt.xlim(-1, 1)
            ax.plot(np.cos(T), np.sin(T), "--", color="black", label="unit circle")
            ax.plot(X, Y, linewidth=2, label="symbol")
            ax.axis("equal")
            ax.legend(loc="best")
            plt.title("Symbol of " + S.name(lambda_bool=True))
        else:
            assert (type(lamb) == np.ndarray and lamb.ndim == 1) or type(lamb) == list
            plt.ylim(-1, 1)
            plt.xlim(-1, 1)
            for l in lamb:
                S = schem(l, order=order)
                X, Y = S.symbol(nparam)
                ax.plot(X, Y, linewidth=2, label=f"$\lambda = $ {l}")
            T = np.linspace(0, 7, 1000)
            ax.plot(np.cos(T), np.sin(T), "--", color="black", label="unit circle")
            ax.axis("equal")
            ax.legend(loc="upper left")
            plt.title("Symbol of " + S.name(lambda_bool=False))

    else:
        if lamb == None:
            lamb = schem(1).CFL
        if isinstance(lamb, (int, float)):
            lamb = np.array([lamb], dtype=float)
        fig = plt.figure(1, figsize=(10, 8))
        plt.title("Symbol of " + schem(1).name(lambda_bool=False))
        ax = fig.add_subplot(111)
        fig.subplots_adjust(left=0.25, bottom=0.25)
        ax.set_xlim(-3, 3)
        ax.axis("equal")

        if len(lamb) == 1:
            lamb0 = lamb[0] / 2
        else:
            lamb0 = (np.min(lamb) + np.max(lamb)) / 2

        Param = np.linspace(0, 1, nparam)
        Param = np.cos(2 * pi * Param) + 1j * np.sin(2 * pi * Param)
        Theta = np.linspace(0, 2 * pi, 500)

        X, Y = schem(lamb0, order=order).symbol(nparam)
        [line] = ax.plot(X, Y, linewidth=2, label="symbol")
        [cible] = ax.plot(np.cos(Theta), np.sin(Theta), "--", color="black", label="unit circle")
        ax.legend(loc="best")
        ax.set_xlim(-6, 2)
        ax.axis("equal")
        if len(lamb) == 1:
            minvalue = 0.0001
            maxvalue = lamb[0]
        else:
            minvalue = np.min(lamb)
            maxvalue = np.max(lamb)

        axlamb = plt.axes([0.25, 0.08, 0.65, 0.03])
        lambda_slider = Slider(
            ax=axlamb,
            label="lambda",
            valmin=minvalue,
            valmax=maxvalue,
            valinit=lamb0,
        )

        def update(val):
            S = schem(lambda_slider.val, order=order)
            X, Y = S.symbol(nparam)
            line.set_xdata(X)
            line.set_ydata(Y)
            fig.canvas.draw_idle()

        # register the update function with each slider
        lambda_slider.on_changed(update)

        # Create a `matplotlib.widgets.Button` to reset the sliders to initial values.
        resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
        button = Button(resetax, "Reset", hovercolor="0.975")

        def reset(event):
            lambda_slider.reset()

        button.on_clicked(reset)
    return ax


def nbrzerosdetKL(schem, left_bound=Dirichlet(), lamb=None, sigma=0, order=2, nparam=100, nlambda=200, nsigma=30, parametrization_bool=True, fig_size=(15, 7), fontsize=18):
    """Computes the number of zeros of the Kreiss-Lopatinskii determinant for different boundary conditions, for different lambdas or/and for different sigmas

    .. seealso:: for more details and examples, see [Boutin, Le Barbenchon, Seguin, 2023 : Stability of finite difference schemes for the hyperbolic initial boundary value problem by winding number computations]

    :param schem: Scheme depending on a parameter lambda
    :type schem: class:`Scheme`
    :param left_bound: Left boundary of the class Boundary, defaults to Dirichlet()
    :type left_bound: class:`Boundary` or list of class:`Boundary`, optional
    :param lamb: A value lambda or a numpy.ndarray of several values of lambda (lambda is the Courant number a.dt/dx), defaults to None
    :type lamb: int or float or list or numpy.ndarray, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: int or float or list or numpy.ndarray, optional
    :param order: Order of the scheme (for Lax-Wendroff scheme), defaults to 2
    :type order: int, optional
    :param nparam: Size of the discretization of the unit circle for the Kreiss-Lopatinskii curve, defaults to 100
    :type nparam: int, optional
    :param nlambda: Size of the discretization of the lambda's, defaults to 200
    :type nlambda: int, optional
    :param nsigma: Size of the discretization of the sigma's, defaults to 30
    :type nsigma: int, optional
    :param parametrization_bool: True to have a refinment close to the origin, False to have a uniform discretization of the unit circle, defaults to True
    :type parametrization_bool: bool, optional
    :param fig_size: Size of the figure, defaults to (15,7)
    :type fig_size: (int,int), optional
    :param fontsize: Size of the font of the text in the plot, defaults to 18
    :type fontsize: int , optional
    :return: the number of zeros of the Kreiss-Lopatinskii determinant for the scheme
    :rtype: plot object
    """
    if lamb is None:
        lambmin = 0.01
        lambmax = schem(1).CFL
    else:
        if isinstance(lamb, (int, float)):
            lamb = np.array([lamb], dtype=float)
        if len(lamb) == 1:
            if lamb[0] <= 0:
                raise ValueError("lambda has to be non negative")
            lambmin = 0.01
            lambmax = lamb[0]
        else:
            lambmin = min(lamb)
            lambmax = max(lamb)
    lambdas = np.linspace(lambmin, lambmax, nlambda)
    if isinstance(left_bound, list) or (isinstance(left_bound, np.ndarray) and left_bound.ndim == 1):
        n = len(left_bound)
        fig, axs = plt.subplots(nrows=n, ncols=1, figsize=fig_size)
        axs[0].set_title(f"Number of zeros of KL determinant for {schem(1/2, order = order, sigma=sigma).name(sigma_bool=True)} with respect to $\lambda$")
        for i, bound in enumerate(left_bound):
            if i == n - 1:
                axs[i].set_xlabel("$\lambda$")
            else:
                axs[i].xaxis.set_ticklabels([])
            axs[i].set_xlim(lambmin, lambmax)
            axs[i].set_ylim(0, 1)
            axs[i].yaxis.set_ticks([])
            axs[i].set_ylabel(bound.name())
            WindingNumbers = []
            for l in lambdas:
                S = schem(l, bound, sigma=sigma, order=order)
                Param, Det = S.detKL(nparam, parametrization_bool)
                WindingNumbers.append(Indice(Det))
            separator = []
            value = []
            for j in range(len(WindingNumbers) - 1):
                if WindingNumbers[j] != WindingNumbers[j + 1]:
                    separator.append((lambdas[j] + lambdas[j + 1]) / 2)
                    value.append(S.r - WindingNumbers[j])
            value.append(WindingNumbers[-1])
            if separator == []:
                axs[i].text((lambmax + lambmin) / 2, 0.5, value[0], horizontalalignment="center", verticalalignment="center", fontsize=fontsize)
            else:
                for j in range(len(separator) + 1):
                    if j == 0:
                        axs[i].plot([separator[j], separator[j]], [0, 1], color="black")
                        axs[i].text(separator[0] / 2, 0.5, value[j], horizontalalignment="center", verticalalignment="center", fontsize=fontsize)
                    elif j == len(separator):
                        axs[i].text((separator[j - 1] + lambmax) / 2, 0.5, value[j], horizontalalignment="center", verticalalignment="center", fontsize=fontsize)
                    else:
                        axs[i].plot([separator[j], separator[j]], [0, 1], color="black")
                        axs[i].text((separator[j] + separator[j - 1]) / 2, 0.5, value[j], horizontalalignment="center", verticalalignment="center", fontsize=fontsize)
        return axs
    else:
        fig, ax = plt.subplots(figsize=fig_size)
        if isinstance(sigma, (int, float)) and not isinstance(sigma, bool):
            WindingNumbers = []
            for l in lambdas:
                S = schem(l, left_bound, sigma=sigma, order=order)
                Param, Det = S.detKL(nparam, parametrization_bool)
                WindingNumbers.append(Indice(Det))
            ax.plot(lambdas, S.r - np.array(WindingNumbers))
            plt.title(
                f"Number of zeros of KL determinant for {schem(1/2,left_bound, order = order, sigma=sigma).name(boundary_bool = True, sigma_bool = True, lambda_bool = False)} with respect to $\lambda$"
            )
        else:
            if sigma:
                sigmin = -1 / 2
                sigmax = 1 / 2
            else:
                sigmax = max(sigma)
                sigmin = min(sigma)

            sigmas = np.linspace(sigmin, sigmax, nsigma)
            WindingNumbers = np.zeros((nsigma, nlambda))
            for i, l in enumerate(lambdas):
                for j, sig in enumerate(sigmas):
                    S = schem(l, left_bound, sigma=sig, order=order)
                    Param, Det = S.detKL(nparam, parametrization_bool)
                    WindingNumbers[j, i] = Indice(Det)
            data = S.r - WindingNumbers
            sigmas, lambdas = np.meshgrid(lambdas, sigmas)
            plt.contourf(sigmas, lambdas, data)
            plt.colorbar()
            plt.title(
                f"Number of zeros of KL determinant for {schem(1/2,left_bound, order = order, sigma=sigma).name(boundary_bool = True, sigma_bool = False, lambda_bool = False)} with respect to $\lambda$ and $\sigma$"
            )
            plt.xlabel("$\lambda$")
            plt.ylabel("$\sigma$")
        return ax



"""
This file aims at testing and plotting the symbol curve
"""

import matplotlib.pyplot as plt
import numpy as np

from boundaryscheme.schemes import *
import boundaryscheme.pyplot as bsplt


# You can import the following different schemes : BeamWarming, Upwind, LaxWendroff, LaxFriedrichs, ThirdOrder, BB, Dissipatif, LW


if __name__ == "__main__":
    """plot the BeamWarming symbol curve with a cursor for lambda between 0 and the CFL condition"""
    bsplt.symbolplot(BeamWarming, lambdacursor=True)

    """plot the BeamWarming symbol curve with a cursor for lambda between 0.7 and 1.8"""
    # bsplt.symbolplot(BeamWarming, [0.7,1.4, 1.8], lambdacursor = True)

    """plot the BeamWarming symbol curve for lambda = 1.4"""
    # bsplt.symbolplot(BeamWarming, 1.4)

    """plot the BeamWarming symbol curves for lambda in [0.7, 1, 1.6]"""
    # bsplt.symbolplot(BeamWarming, np.array([0.7, 1, 1.6]))

    plt.show()



"""
This file aims at testing and plotting the Kreiss-Lopatinskii determinant curve
"""

import matplotlib.pyplot as plt
import numpy as np

from boundaryscheme.boundaries import *
from boundaryscheme.schemes import *
import boundaryscheme.pyplot as bsplt


# You can import the following different schemes : BeamWarming, Upwind, LaxWendroff, LaxFriedrichs, ThirdOrder, BB, Dissipatif, LW

# You can import the following different boundaries : Dirichlet, SILW, DDJ


if __name__ == "__main__":
    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming and for lambda = 1.4"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lamb = 1.4)

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming and for lambda in [1.4, 0.7, 1] with sigma = 0.4"""
    bsplt.detKLplot(BeamWarming, SILW(2, 3), lamb=np.array([1.4, 0.7, 1]), sigma=0)

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming and for lambda = 1 with sigma in [-0.2, 0, 0.1,0.4]"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lamb = 1, sigma = np.array([-0.2, 0, 0.1,0.4]))

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming and for lambda in [1.4, 0.7] with sigma in [-0.2, 0, 0.4]"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lamb = [1.4, 0.7], sigma = np.array([-0.2, 0, 0.4]))

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming and for lambda = 1.4 and with a cursor for sigma between -0.5 and 0.5"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lamb = 1.4, sigmacursor = True)

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming with a cursor for lambda between 0 and the CFL condition"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lambdacursor = True)

    """plot the Kreiss-Lopatinskii determinant curve for BeamWarming with a cursor for lambda between 0 and the CFL condition and a cursor for sigma between -0.5 and 0.5"""
    # bsplt.detKLplot(BeamWarming, SILW(2,3), lambdacursor = True, sigmacursor = True)

    plt.show()



"""
This file defines the schemes classes
"""

import copy
import sympy as sp
import numpy as np
from math import *
from cmath import *
import numpy.polynomial.polynomial as nppol

from .utils import boundary_to_boundary_quasi_toep, sort, epsilon, eta_func, parametrization, Lwconstructor
from .boundaries import Dirichlet


class Scheme:
    """This is a class to represent a numerical scheme.

    :param inter: List of float coefficient that represent the interior scheme
    :type inter: list
    :param center: Index of the central coefficient of the scheme
    :type center: int
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional

    .. seealso:: [Boutin, Le Barbenchon, Seguin, 2023 : Stability of finite difference schemes for the hyperbolic initial boundary value problem by winding number computations]
    """

    def __init__(self, inter, center, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.inter = inter
        self.center = center
        self.r = center
        self.p = len(self.inter) - self.r - 1
        self.boundaryname = boundary.name()
        littleboundary, self.gn = boundary(center, sigma=self.sigma)
        self.m = max(len(littleboundary[0]), self.r + self.p)
        self.boundary = np.zeros((self.r, self.m))
        self.boundary[: self.r, : len(littleboundary[0])] = littleboundary
        self.boundary_quasi_toep, self.bn = boundary_to_boundary_quasi_toep(self.boundary, self.gn, inter, center)

        z0 = sp.Symbol("z0", imaginary=True)
        B = copy.deepcopy(self.boundary_quasi_toep)
        B = sp.Matrix(B) - z0 * sp.eye(self.r, self.m)
        b = sp.ones(1, self.r + 1)
        for i in range(self.r + 1):
            b[i] = sp.Symbol("b" + str(i), imaginary=True)
        for j in range(self.m - self.r):
            row = sp.zeros(1, self.m)
            for k in range(self.r + 1):
                row[self.m - self.r - j + k - 1] = b[k]
            B = B - np.dot(B[:, self.m - 1 - j], row)
        self.DKL = sp.lambdify([z0, b], sp.det(B[: self.r, : self.r]), "numpy")

    def scheme(self):
        """Returns the parameters of the scheme : the interior coefficients et the center
        :return: Parameters of the scheme
        :rtype: (list,int)
        """
        return self.inter, self.center

    def toep(self, J, right_bound=np.array([[]])):
        """Returns the Toeplitz matrix of the scheme with two boundaries
        :param J: Size of the Toeplitz matrix
        :type J: int
        :param right_bound: Right boundary of the scheme
        :type right_bound: class:`Boundary`
        :return: Toeplitz matrix of the scheme
        :rtype: numpy.ndarray
        """
        A = np.zeros((J, J))
        for k in range(len(self.inter)):
            A += self.inter[k] * np.diag(np.ones(J - abs(k - self.center)), k - self.center)
        A[0 : len(self.boundary_quasi_toep), 0 : len(self.boundary_quasi_toep[0])] = self.boundary_quasi_toep
        A[J - len(right_bound) : J, J - len(right_bound[0]) : J] = right_bound
        return A

    def toep_circ(self, J):
        """Returns the circulant Toeplitz matrix of the scheme
        :param J: Size of the circulant matrix
        :type J: int
        :return: Circulant matrix of the scheme
        :rtype: numpy.ndarray
        """
        C = np.zeros((J, J))
        for k in range(len(self.inter)):
            C += self.inter[k] * np.diag(np.ones(J - abs(k - self.center)), k - self.center)
            if k - self.center >= 0:
                C += self.inter[k] * np.diag(np.ones(abs(k - self.center)), -J + k - self.center)
            else:
                C += self.inter[k] * np.diag(np.ones(abs(k - self.center)), J + k - self.center)
        return np.matrix(C)

    def symbol(self, n=300):
        """To draw the symbol curve of the scheme
        :param n: Number of points of the parametrisation of the unit circle, defaults to 300
        :type n: int, optional
        :return: Abscissa and Ordinate of the symbol curve of the scheme
        :rtype: (list,list)
        """
        T = np.linspace(0, 7, n)
        X = np.zeros_like(T)
        Y = np.zeros_like(T)
        for i in range(len(self.inter)):
            X += self.inter[i] * np.cos((i - self.center) * T)
            Y += self.inter[i] * np.sin((i - self.center) * T)
        return X, Y

    def pol(self, z):
        """Returns the characteristic polynomial of the scheme
        :param z: Parameter z of the characteristic polynomial
        :type z: complex
        :return: Characteristic polynomial of the scheme
        :rtype: class: Polynomial
        """
        r = self.center
        monome = nppol.Polynomial([0, 1])
        P = nppol.Polynomial(self.inter) - z * monome**r
        return P

    def roots(self, z):
        """Returns the roots of the characteristic polynomial
        :param z: Parameter z of the characteristic polynomial
        :type z: complex
        :return: Roots of the characteristic polynomial sorting by utils.sort function
        :rtype: list
        """
        Racinestotales = self.pol(z).roots()
        return sort(Racinestotales)

    def count_root(self, eta, eps, z0, kappa):
        """Counts the number of roots from kappa(z0) that comes from the inside of the unit disk
        :param eta: Gap between z and z0
        :type eta: float
        :param eps: Threshold
        :type eps: float
        :param z0: Parameter z0
        :type z0: complex
        :param kappa: Multiple root kappa linked to z0
        :type kappa: complex
        :return: Number of roots coming from the inside of the unit disk
        :rtype: int
        """
        z = z0 + eta * z0 / (2 * abs(z0))
        NewRoots = self.roots(z)
        selection = list(filter(lambda k: abs(k - kappa) < eps, NewRoots))
        return len(list(filter(lambda k: abs(k) < 1, selection)))

    def Kappa(self, z0):
        """Selection of kappas that come from the inside of the unit disk
        :param z0: Parameter z0 of the characteristic polynomial
        :type z0: complex
        :return: Roots of the characteristic polynomial for z0 that are inside the unit disk or are coming from the inside of the unit disk
        :rtype: list
        """
        delta = 10 ** (-10)
        Racinestotales = self.roots(z0)
        eps = epsilon(Racinestotales)
        RootsFromInside = []
        stock = []
        for x in Racinestotales:
            if abs(x) < 1 - delta:
                RootsFromInside.append(x)
            elif abs(x) < 1 + delta and x not in stock:
                stock.append(x)
                eta = eta_func(eps, x, 1000, self.pol(z0), self.center)
                n = self.count_root(eta, eps, z0, x)
                for i in range(n):
                    RootsFromInside.append(x)
        assert len(RootsFromInside) == self.center
        return RootsFromInside

    def detKL(self, n_param, parametrization_bool):
        """Computes the Kreiss-Lopatinskii determinant on the unit circle
        :param n_param: Number of discretization of the unit circle
        :type n_param: int
        :param parametrization_bool: If True, the discretization is refined close to the origin. If False, the discretization is uniform on the unit circle.
        :type parametrization_bool: bool
        :return: The parametrization of the unit circle and the value of the Kreiss-Lopatinskii determinant on the unit circle
        :rtype: (numpy.ndarray, numpy.ndarray)
        """

        def scalar_detKL(self, z):
            Rz = nppol.polyfromroots(self.Kappa(z))
            return self.DKL(z, Rz)

        if parametrization_bool:
            return parametrization(n_param, lambda z: scalar_detKL(self, z))
        else:
            param = np.linspace(0, 2 * pi, n_param)
            return param, np.vectorize(scalar_detKL)(self, np.exp(1j * param))

    def shortname(self):
        """Gives a short name of the scheme"""
        r = self.center
        s = "U_j^{n+1} = " + str(round(self.inter[0], 3)) + " U_{j-" + str(r) + "}^n + "
        for i in range(1, len(self.inter) - 1):
            if r > i:
                s += str(round(self.inter[i], 3)) + " U_{j-" + str(r - i) + "}^n + "
            elif r == i:
                s += str(round(self.inter[i], 3)) + " U_j^n + "
            else:
                s += str(round(self.inter[i], 3)) + " U_{j+" + str(i - r) + "}^n + "
        if r > len(self.inter) - 1:
            s += str(round(self.inter[-1], 3)) + " U_{j-" + str(r - i) + "}^n"
        elif r == len(self.inter) - 1:
            s += str(round(self.inter[-1], 3)) + " U_j^n"
        else:
            s += str(round(self.inter[-1], 3)) + " U_{j+" + str(i - r + 1) + "}^n"
        return s

    def name(self, boundary_bool=False, sigma_bool=False, lambda_bool=False):
        """Gives a name of the scheme"""
        string = self.shortname()
        if boundary_bool:
            string += r" with boundary " + self.boundaryname
        if sigma_bool and self.sigma != 0:
            string += r" with $\sigma$ = " + str(self.sigma)
        if lambda_bool:
            string += r" for $\lambda$ = " + str(round(self.lamb, 2))
        return string

    def __repr__(self):
        """Print the numerical scheme"""
        r = self.center
        s = "the interior equation of the scheme is :\n" + self.shortname() + "\nthe boundary equation is :"
        for i in range(len(self.boundary)):
            s += "\nU_{-" + str(r - i) + "}^n = "
            for j in range(len(self.boundary[i]) - 1):
                s += str(round(self.boundary[i, j], 3)) + " U_" + str(j) + "^n + "
            s += str(round(self.boundary[i, -1], 3)) + " U_" + str(len(self.boundary[0]) - 1) + "^n"
        return s


class SchemeP0(Scheme):
    """This is a class to represent a totally upwind scheme (with p = 0).

    :param inter: List of float coefficient that represent the interior scheme
    :type inter: list
    :param center: Index of the central coefficient of the scheme
    :type center: int
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional

    .. seealso:: [Boutin, Le Barbenchon, Seguin, 2023 : On the stability of totally upwind schemes for the hyperbolic initial boundary value problem]
    """

    def __init__(self, inter, center, boundary, sigma=0):
        """Constructor Method"""
        super().__init__(inter=inter, center=center, boundary=boundary, sigma=sigma)
        assert len(inter) == center + 1

    def detKL(self, n_param, parametrization_bool):
        """
        Redefinition of the detKL method of the :class: `Scheme` but in the particular case of totally upwind scheme
        :param n_param: Number of discretization of the unit circle
        :type n_param: int
        :param parametrization_bool: If True, the discretization is refined close to the origin. If False, the discretization is uniform on the unit circle.
        :type parametrization_bool: bool
        :return: The parametrization of the unit circle and the value of the Kreiss-Lopatinskii determinant on the unit circle
        :rtype: (numpy.ndarray, numpy.ndarray)
        """

        def scalar_detKL(self, z):
            b = np.array([self.inter[i] / (self.inter[-1] - z) for i in range(len(self.inter))])
            b[-1] = 1
            return self.DKL(z, b)

        if parametrization_bool:
            return parametrization(n_param, lambda z: scalar_detKL(self, z))
        else:
            param = np.linspace(0, 2 * pi, n_param)
            return param, np.vectorize(scalar_detKL)(self, np.exp(1j * param))


class BeamWarming(SchemeP0):
    """This is a class to represent the Beam-Warming scheme (which is totally upwind).

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        self.p = 0
        self.r = 2
        coeff1 = (lamb - 1) * (lamb - 2) / 2
        coeff2 = lamb * (2 - lamb)
        coeff3 = lamb * (lamb - 1) / 2
        self.inter = np.array([coeff3, coeff2, coeff1])
        self.center = 2
        self.CFL = 2
        while self.inter[0] == 0:
            self.inter = self.inter[1:]
            self.center -= 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma)

    def shortname(self):
        """Name method"""
        return "BeamWarming"


class Upwind(SchemeP0):
    """This is a class to represent the Upwind scheme (which is totally upwind).

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        self.p = 0
        self.r = 1
        coeff1 = 1 - lamb
        coeff2 = lamb
        self.inter = np.array([coeff2, coeff1])
        self.center = 1
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "Upwind"


class LaxWendroff(Scheme):
    """This is a class to represent the Lax-Wendroff scheme (of order 2).

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        coeff1 = -(lamb - lamb**2) / 2
        coeff2 = 1 - lamb**2
        coeff3 = (lamb**2 + lamb) / 2
        self.inter = np.array([coeff3, coeff2, coeff1])
        self.center = 1
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "LaxWendroff"


class LaxFriedrichs(Scheme):
    """This is a class to represent the Lax-Friedrichs scheme.

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        self.inter = np.array([(1 + lamb) / 2, 0, (1 - lamb) / 2])
        self.center = 1
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "LaxFriedrichs"


class ThirdOrder(Scheme):
    """This is a class to represent the O3-scheme.

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        self.r = 2
        self.p = 1
        self.inter = np.array(
            [
                lamb**3 / 6 - lamb / 6,
                lamb + lamb * lamb / 2 - lamb**3 / 2,
                1 - lamb / 2 - lamb * lamb + lamb**3 / 2,
                -(lamb**3) / 6 + lamb * lamb / 2 - lamb / 3,
            ]
        )
        self.center = 2
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "ThirdOrder"


class BB(Scheme):
    """This is a class to represent an example of numerical scheme.

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.lamb = lamb
        self.inter = np.array([self.lamb / 4, self.lamb / 4, 1 - self.lamb / 4, -self.lamb / 4])
        self.center = 2
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "BB"


class Dissipatif(Scheme):
    """This is a class to represent a dissipative scheme.

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    :param D: A parameter of the dissipativity, defaults to 0
    :type D: float, optional
    """

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, D=0, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.D = D
        self.lamb = lamb
        l = 2
        self.inter = np.array([lamb / (2 * l), self.D, 1 - 2 * self.D, self.D, -lamb / (2 * l)])
        self.center = 2
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return "Dissipative"


class LW(Scheme):
    """This is a class to represent a Lax-Wendroff scheme of any order.

    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :param boundary: Boundary condition, defaults to Dirichlet()
    :type boundary: class:`Boundary`, optional
    :param sigma: Gap between the mesh and the boundary condition, defaults to 0
    :type sigma: float, optional
    :param order: Order of the Lax-Wendroff scheme, defaults to 2
    :type order: int, optional
    """

    allordersschem = [None, None, Lwconstructor(2)]

    def __init__(self, lamb, boundary=Dirichlet(), sigma=0, order=2, **kwargs):
        """Constructor method"""
        self.sigma = sigma
        self.order = order
        self.lamb = lamb
        n = len(LW.allordersschem) - 1
        if order > n:
            for i in range(order - n):
                LW.allordersschem.append(None)
            LW.allordersschem[order] = Lwconstructor(order)
        if LW.allordersschem[order] == None:
            LW.allordersschem[order] = Lwconstructor(order)
        self.inter, self.center = LW.allordersschem[order](lamb)
        self.CFL = 1
        super().__init__(inter=self.inter, center=self.center, boundary=boundary, sigma=sigma, **kwargs)

    def shortname(self):
        """Name method"""
        return f"Lax Wendroff {self.order}"


# class LeapFrog(Scheme):
#     raise NotImplementedError('')


# class CrankNicholson(Scheme):
#     raise NotImplementedError('')



"""
This file implements a routine to perform the "point in polygon" inclusion test
"""

###### Complex winding number ############

#!/usr/bin/env python
#
# routine for performing the "point in polygon" inclusion test

# Copyright 2001, softSurfer (www.softsurfer.com)
# This code may be freely used and modified for any purpose
# providing that this copyright notice is included with it.
# SoftSurfer makes no warranty for this code, and cannot be held
# liable for any real or imagined damage resulting from its use.
# Users of this code must verify correctness for their application.

# translated to Python by Maciej Kalisiak <mac@dgp.toronto.edu>

#   a Point is represented as a tuple: (x,y)

# ===================================================================

# is_left(): tests if a point is Left|On|Right of an infinite line.

#   Input: three points P0, P1, and P2
#   Return: >0 for P2 left of the line through P0 and P1
#           =0 for P2 on the line
#           <0 for P2 right of the line
#   See: the January 2001 Algorithm "Area of 2D and 3D Triangles and Polygons"


def is_left(P0, P1, P2):
    return (P1[0] - P0[0]) * (P2[1] - P0[1]) - (P2[0] - P0[0]) * (P1[1] - P0[1])


# ===================================================================

# wn_PnPoly(): winding number test for a point in a polygon
#     Input:  P = a point,
#             V[] = vertex points of a polygon
#     Return: wn = the winding number (=0 only if P is outside V[])


def wn_PnPoly(P, V):
    wn = 0  # the winding number counter

    # repeat the first vertex at end
    V = tuple(V[:]) + (V[0],)

    # loop through all edges of the polygon
    for i in range(len(V) - 1):  # edge from V[i] to V[i+1]
        if V[i][1] <= P[1]:  # start y <= P[1]
            if V[i + 1][1] > P[1]:  # an upward crossing
                if is_left(V[i], V[i + 1], P) > 0:  # P left of edge
                    wn += 1  # have a valid up intersect
        else:  # start y > P[1] (no test needed)
            if V[i + 1][1] <= P[1]:  # a downward crossing
                if is_left(V[i], V[i + 1], P) < 0:  # P right of edge
                    wn -= 1  # have a valid down intersect
    return wn


def translate(L):
    """
    take a list of complex values
    return the list of the coordinates of each complex in the R^2 plan
    """
    V = []
    for z in L:
        V.append((z.real, z.imag))
    return V


def Indice(L):
    """
    take a list of (x_i,x_j) representing a polygon
    return the winding number of the origin with respect to the curve
    """
    return wn_PnPoly((0, 0), translate(L))



"""
This file defines the boundaries classes
"""

import scipy as sc
import numpy as np
from math import *

from .utils import coefBinomial


class Boundary:
    """This is a class to represent the boundary condition."""

    def __init__(self, B):
        self.B = B

    def __call__(self, r):
        return self.B


class Dirichlet(Boundary):
    """This is a class to represent the homogeneous dirichlet boundary condition.

    :param d: Order of consistency
    :type d: int
    """

    def __init__(self):
        self.d = 0

    def __call__(self, r, sigma=0):
        """Builds the left boundary condition

        :param r: Number of left ghost points to define the numerical scheme considered
        :type r: int
        :return: The boundary condition B and a function gn such as (U_{-r},...,U_{-1}) = B (U_0,...,U_{m-1}) + gn(t)
        :rtype: (numpy.ndarray,function)
        """
        return np.zeros((r, 1)), lambda t: 0

    def name(self):
        """Name"""
        return "Dirichlet"


class SILW(Boundary):
    """
    This is a class to represent the Simplified Inverse Lax Wendroff boundary procedure, see [Vilar,Shu, 2015 : Development and stability analysis of the inverse Lax Wendroff boundary treatment for central compact schemes]

    :param kd: Index of the beginning of extrapolation using
    :type kd: int
    :param d: Order of consistency of the boundary condition
    :type d: int
    :param dx: Space discretization, defaults to 1/10
    :type dx: float, optional
    :param a: Velocity, defaults to 1
    :type a: float, optional
    :param gderivatives: List of the boundary data and its derivatives, defaults to None
    :type gderivatives: list, optional
    """

    def __init__(self, kd, d, dx=0.1, a=1, gderivatives=None):
        """Constructor method"""
        self.kd = kd
        self.d = d
        self.dx = dx
        self.a = a
        if gderivatives == None:
            gderivatives = [lambda x: 0 for i in range(kd)]
        if len(gderivatives) < kd:
            raise IndexError("gderivatives needs at least kd functions")
        self.gderivatives = gderivatives

    def __call__(self, r, sigma=0):
        """Builds the left boundary condition

        .. note:: m is the number of points needed to define the boundary condition, it is equal to d here.

        :param r: Number of left ghost points to define the numerical scheme considered
        :type r: int
        :param sigma: Gap between the mesh and the boundary condition, defaults to 0
        :type sigma: float, optional
        :return: The boundary condition B and a function gn such as (U_{-r},...,U_{-1}) = B (U_0,...,U_{m-1}) + gn(t)
        :rtype: (numpy.ndarray,function)
        """
        B = np.zeros((r, self.d))
        for j in range(r):
            for m in range(self.kd, self.d):
                for l in range(m + 1):
                    B[r - 1 - j, l] += (-(j + 1) + sigma) ** m / factorial(m) * coefBinomial(m, l) * (-1) ** (m - l)

        def gn_func(t):
            gn = np.zeros(r)
            for j in range(r):
                for k in range(self.kd):
                    gn[r - 1 - j] += (-(j + 1) + sigma) ** k / factorial(k) * self.dx**k * (-1) ** k * self.gderivatves[k](t) / (self.a**k)
            return gn

        return B, gn_func

    def name(self):
        """Name"""
        return f"S{self.kd}ILW{self.d}"


class DDJ(Boundary):
    """
    This is a class to represent the boundary Reconstruction procedure explained in [Boutin, Le Barbenchon, Seguin, 2023 : Stability of finite difference schemes for the hyperbolic initial boundary value problem by winding number computations] and in [Dakin Desprès Jaouen, 2018 : Inverse Lax–Wendroff boundary treatment for compressible Lagrange-remap hydrodynamics on cartesian grids]

    .. warning:: Warning d represents the order and kd represente the truncature, the notation is not in the same order than SILW, it is to respect the order of [DakinDespresJaouen18]

    :param kd: Index of the beginning of extrapolation using
    :type kd: int
    :param d: Order of consistency of the boundary condition
    :type d: int
    :param dx: Space discretization, defaults to 1/10
    :type dx: float, optional
    :param a: Velocity, defaults to 1
    :type a: float, optional
    :param gderivatives: List of the boundary data and its derivatives, defaults to None
    :type gderivatives: list, optional
    """

    def __init__(self, d, kd, dx=0.1, a=1, gderivatives=None):
        """Constructor method"""
        self.d = d
        self.kd = kd
        self.dx = dx
        self.a = a
        if gderivatives == None:
            gderivatives = [lambda x: 0 for i in range(kd)]
        if len(gderivatives) < kd:
            raise IndexError("gderivatives needs at least kd functions")  # à vérifier
        self.gderivatives = gderivatives

    def __call__(self, r, sigma=0):
        """Builds the left boundary condition

        .. note:: m is the number of points needed to define the boundary condition, it is equal to d here.

        :param r: Number of left ghost points to define the numerical scheme considered
        :type r: int
        :param sigma: Gap between the mesh and the boundary condition, defaults to 0
        :type sigma: float, optional
        :return: The boundary condition B and a function gn such as (U_{-r},...,U_{-1}) = B (U_0,...,U_{m-1}) + gn(t)
        :rtype: (numpy.ndarray,function)
        """
        ymoins = np.zeros((r, self.d - self.kd - 1))
        yplus = np.zeros((self.d - self.kd - 1, self.d - self.kd - 1))
        for i in range(1, r + 1):
            for j in range(1, self.d - self.kd):
                ymoins[i - 1, j - 1] = ((1 - i + 1 / 2 - sigma) ** (self.kd + j + 1) - (1 - i - 1 / 2 - sigma) ** (self.kd + j + 1)) / (factorial(self.kd + j + 1))
        for i in range(1, self.d - self.kd):
            for j in range(1, self.d - self.kd):
                yplus[i - 1, j - 1] = ((i + 1 / 2 - sigma) ** (self.kd + j + 1) - (i - 1 / 2 - sigma) ** (self.kd + j + 1)) / (factorial(self.kd + j + 1))
        if np.linalg.det(yplus) == 0:
            print(yplus, sigma)
        ConditionBord = ymoins.dot(np.linalg.inv(yplus))
        # """write the boundary condition with gn"""
        return ConditionBord[::-1], lambda x: 0

    def name(self):
        """Name"""
        return f"DDJ{self.d},{self.kd}"



"""
This file aims at testing and plotting the number of zeros of the Kreiss-Lopatinskii determinant
"""

import matplotlib.pyplot as plt
import numpy as np

from boundaryscheme.schemes import *
from boundaryscheme.boundaries import *
import boundaryscheme.pyplot as bsplt


# You can import the following different schemes : BeamWarming, Upwind, LaxWendroff, LaxFriedrichs, ThirdOrder, BB, Dissipatif, LW

# You can import the following different boundaries : Dirichlet, SILW, DDJ


if __name__ == "__main__":
    """plot the number of zeros of Kreiss-Lopatinskii determinant of BeamWarming with SILW(2,3) for lambda between 0 and the CFL condition"""
    # bsplt.nbrzerosdetKL(BeamWarming, SILW(2,3))

    """plot the number of zeros of Kreiss-Lopatinskii determinant of BeamWarming with SILW(2,3) for lambda between 1.4 and 1.7"""
    # bsplt.nbrzerosdetKL(BeamWarming,  SILW(2,3),lamb = np.array([1.4,1.7]), nparam = 150)

    """plot the number of zeros of Kreiss-Lopatinskii determinant of BeamWarming with several boundary conditions"""
    bsplt.nbrzerosdetKL(BeamWarming, [SILW(1, 3), SILW(2, 3), SILW(2, 4), SILW(3, 4), DDJ(3, 0), DDJ(3, 1)], nlambda=60)

    """plot the number of zeros of Kreiss-Lopatinskii determinant of BeamWarming with SILW(2,3) with respect to lambda and sigma"""
    # bsplt.nbrzerosdetKL(BeamWarming, SILW(2,3), sigma=True, nlambda=40, nsigma = 20)

    plt.show()



import os
import re
from setuptools import setup, find_packages


def rel_file(*args):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), *args)


def read_from(filename):
    with open(filename) as fp:
        return fp.read()


def get_version():
    data = read_from(rel_file('boundaryscheme', '__init__.py'))
    return re.search(r"__version__ = '([^']+)'", data).group(1)


def get_requirements():
    data = read_from(rel_file('requirements.txt'))
    lines = map(lambda s: s.strip(), data.splitlines())
    return list(filter(None, lines))


setup(
    name='boundaryscheme',
    version=get_version(),
    description="This library implements Kreiss-Lopatinskii determinant for numerical scheme with boundary",
    long_description=read_from('README.md'),
    long_description_content_type='text/markdown',
    author='Pierre Le Barbenchon',
    packages=find_packages(),
    install_requires=get_requirements(),
)



"""
This file defines utils function specific to the library
"""

import numpy as np
from math import *
from cmath import *


def coefBinomial(n, k):
    """Computes the binomial coefficient

    :param n: Number of elements
    :type n: int
    :param k: Number of choosen elements
    :type k: int
    :return: Binomial coefficient "n choose k"
    :rtype: int
    """
    if k > (n - k):
        k = n - k
    coef = 1
    for i in range(k):
        coef = coef * (n - i)
        coef = coef // (i + 1)
    return coef


def sort(L):
    """Sort (bubble sort)

    :param L: List of complex numbers
    :type L: list
    :return: The same list but in the ascending order of modulus and for values with same modulus, sort in the ascending order of argument (between -pi and pi)
    :rtype: list
    """
    nbr = len(L)
    for i in range(nbr):
        for j in range(nbr - 1):
            if abs(L[j + 1]) < abs(L[j]):
                x = L[j + 1]
                L[j + 1] = L[j]
                L[j] = x
    for i in range(nbr):
        for j in range(nbr - 1):
            if abs(L[j + 1]) == abs(L[j]):
                if phase(L[j + 1]) < phase(L[j]):
                    x = L[j + 1]
                    L[j + 1] = L[j]
                    L[j] = x
    return L


def sector(z):
    """Finds the sector of z

    .. warning:: the argument is between -pi and pi and here "a" is between 0 and 7

    :param z: Complex number
    :type z: complex
    :return: The sector of z, it is an integer "a" between 0 and 7 which correspond to the angular sector [a*pi/4, (a+1)*pi/4[
    :rtype: int
    """
    ph = phase(z) * 4 / pi
    if ph < 0:
        return int(ph) + 7
    else:
        return int(ph)


def neighboor(sector1, z2):
    """Checks if z2 is in a neighboring sector of sector1

    :param sector1: Integer between 0 and 7 and represent the angular sector [sector1*pi/4, (sector1+1)*pi/4[
    :type sector1: int
    :param z2: Complex number
    :type z2: complex
    :return: True if and only if the complex z2 is in a neighboring sector of [sector1*pi/4, (sector1+1)*pi/4[
    :rtype: bool
    """
    return abs((sector(z2) - sector1) % 8) <= 1


def parametrization(n_param, curve_formula):
    """Finds the discretization of the curve refine if it is needed (as [ZapataMartin2014] procedure).

    :param n_param: Integer for the default discretization of [0,2pi]
    :type n_param: int
    :param curve_formula: A fonction : z in the unit circle mapsto a complex and represent a curve
    :type curve_fomula: function
    :return: The discretization of the unit circle and the values of the curve for this discretization
    :rtype: (list,list)
    """
    dx = 2 * pi / n_param
    current_dx = dx
    current_param = 0
    Param = [current_param]
    curve = [curve_formula(1)]
    current_sector = sector(curve_formula(np.exp(1j * Param[0])))  # calcul le secteur du premier point
    c = 0
    s = 0
    while Param[-1] < 2 * pi:
        current_param = Param[-1] + current_dx
        current_curve_point = curve_formula(np.exp(1j * current_param))
        if neighboor(current_sector, current_curve_point):
            Param.append(current_param)
            curve.append(current_curve_point)
            current_sector = sector(current_curve_point)
            current_dx = dx
            c = 0
        elif c < 40:
            current_dx = current_dx / 2
            c += 1
        else:
            current_param = Param[-1] + dx
            Param.append(current_param)
            curve.append(curve_formula(np.exp(1j * current_param)))
            current_sector = sector(curve[-1])
            current_dx = dx
            c = 0
        s += 1
    if Param[-1] > 2 * pi:
        Param[-1] = 0
        curve[-1] = curve_formula(1)
    return np.array(Param), np.array(curve)


def epsilon(L):
    """Computes the half of the smallest distance between two distincts elements of the list L

    :param L: List of complex numbers (z_i)
    :type L: list
    :return: min \|z_i-z_j| / 2 (when z_i != z_j)
    :rtype: float
    """
    assert len(L) > 1
    diff = []
    for i in range(len(L)):
        for j in range(i + 1, len(L)):
            if L[i] != L[j]:
                diff.append(abs(L[i] - L[j]))
    mini = min(diff)
    assert mini > 0
    return mini / 2


def eta_func(eps, kappa0, N, polynom, r):
    """Computes the distance eta
    .. note:: for more details, see [Boutin, Le Barbenchon, Seguin, 2023 : Stability of finite difference schemes for the hyperbolic initial boundary value problem by winding number computations]

    :param eps: Radius
    :type eps: float
    :param kappa0: Center
    :type kappa0: complex
    :param N: Number of discretization of the circle centered in kappa0 of radius eps
    :type N: int
    :param polynom: Polynomial
    :type polynom: Polynomial
    :param r: Integer
    :type r: int
    :return: min \|polynom(kappa)|/(1+eps)^r for kappa on the circle centered in kappa0 of radius eps
    :rtype: float
    """
    theta = np.linspace(0, 2 * pi, N)
    circle = np.cos(theta) + 1j * np.sin(theta)
    kappas = kappa0 + eps * circle
    val_pol = np.abs(polynom(kappas))
    mini = min(val_pol)
    assert mini > 0
    return mini / ((1 + eps) ** r)


def schemsum(schem1, schem2):
    """Addition of two numerical schemes

    :param schem1: Representation of a numerical scheme with the interior coefficients and the center
    :type schem1: (list,int)
    :param schem2: Representation of a numerical scheme with the interior coefficients and the center
    :type schem2: (list,int)
    :return: Sum of schem1 and schem2
    :rtype: (list,int)
    """
    inter1, center1 = schem1
    inter2, center2 = schem2
    center = max(center1, center2)
    p = max(len(inter1) - center1, len(inter2) - center2) - 1
    newinter = np.zeros(p + center + 1)
    for i in range(len(inter1)):
        newinter[center - center1 + i] += inter1[i]
    for i in range(len(inter2)):
        newinter[center - center2 + i] += inter2[i]
    return newinter, center


def schemscal(schem, scal):
    """Multiplication of a scheme by a scalar

    :param schem: Representation of a numerical scheme with the interior coefficients and the center
    :type schem: (list,int)
    :param scal: Scalar element
    :type scal: float
    :return: Multiplication of schem by scal
    :rtype: (list,int)
    """
    inter, center = schem
    return scal * inter, center


def boundary_to_boundary_quasi_toep(boundary_condition, gn, inter, center):
    """Transformation from a boundary condition written with ghost points to a boundary condition written as the extraction of the r first rows of the Quasi-Toeplitz matrix

    :param boundary_condition: Boundary condition written as U_{-r} = ..., ...,U_{-1} =..., etc
    :type boundary_condition: numpy.ndarray
    :param gn: Boundary data
    :type gn: function
    :param inter: List of float coefficient that represent the interior scheme
    :type inter: list
    :param center: Index of the central coefficient of the scheme
    :type center: int
    :return: Extraction of the r first rows of the Quasi-Toeplitz matrix and the function b_n(t) such that (U_0^{n+1},...,U_{r-1}^{n+1}) = T_J (U_0^n,...,U_{m-1}^n) + b_n(t^n)
    :rtype: (numpy.ndarray,function)
    """
    r = center
    p = len(inter) - r - 1
    m = max(len(boundary_condition[0]), r + p)
    A = np.zeros((r, r))
    for j in range(r):
        A += inter[j] * np.diag(np.ones(r - j), j)
    Ap = np.zeros((r, m))
    for j in range(1, r + p + 1):
        if j < r:
            Ap[:r, :r] += inter[j] * np.diag(np.ones(j), -r + j)
        else:
            Ap[:r, j - r : j] += inter[j] * np.diag(np.ones(r), 0)
    BB = np.zeros((r, m))
    BB[:r, : len(boundary_condition[0])] = boundary_condition

    def bn_func(t):
        return A.dot(gn(t))

    return A.dot(BB) + Ap, bn_func


def boundary_quasi_toep_to_boundary(quasi_toep_boundary, bn, inter, center):
    """Transformation from a boundary condition written as the extraction of the r first rows of the Quasi-Toeplitz matrix to a boundary condition written with ghost points.
    :param quasi_toep_boundary: Boundary condition BB written as the extraction of the r first rows of a Quasi-Toeplitz matrix
    :type quasi_toep_boundary: numpy.ndarray
    :param bn: Boundary data bn such that (U_0^{n+1},...,U_{r-1}^{n+1}) = BB (U_0^n,...,U_{m-1}^n) + b_n(t^n)
    :type bn: function
    :param inter: List of float coefficient that represent the interior scheme
    :type inter: list
    :param center: Index of the central coefficient of the scheme
    :type center: int
    :return: Boundary condition B written as U_{-r} = ..., ..., U_{-1} =...
    and the function g_n such that (U_{-r}^n,...,U_{-1}^n) = B(U_0^n,...,U_{m-1}^n) + g_n(t^n)
    :rtype: (numpy.ndarray,function)
    """
    r = center
    p = len(inter) - r - 1
    m = max(len(quasi_toep_boundary[0]), r + p)
    A = np.zeros((r, r))
    for j in range(r):
        A += inter[j] * np.diag(np.ones(r - j), j)
    Ap = np.zeros((r, m))
    for j in range(1, r + p + 1):
        if j < r:
            Ap[:r, :r] += inter[j] * np.diag(np.ones(j), -r + j)
        else:
            Ap[:r, j - r : j] += inter[j] * np.diag(np.ones(r), 0)
    BB = np.zeros((r, m))
    BB[:r, : len(quasi_toep_boundary[0])] = quasi_toep_boundary

    def gn_func(t):
        return np.linalg.inv(A).dot(bn(t))

    return np.linalg.inv(A).dot(BB - Ap), gn_func


def recflux(N, lamb):
    """Computes the coefficient of the Lax-Wendroff scheme at any order. Recursive function.

    :param N: Order of Lax Wendroff scheme
    :type N: int
    :param lamb: The Courant number, i.e  a.dt/dx where "a" is the velocity, "dt" the time discretization and "dx" the space discretization
    :type lamb: float
    :return: Parameters of the Lax-Wendroff scheme : interior coefficient and the center.
    :rtype: (np.ndarray,int)
    """
    if N == 1:
        return np.array([1]), 0
    else:
        m = N // 2
        M = (N - 1) // 2
        a = np.zeros(N)
        for k in range(N):
            a[N - 1 - k] = (-1) ** (k + N) * coefBinomial(N - 1, k)
        center = N - 1 - m
        prod = 1
        for i in range(-m, M + 1):
            if i != 0:
                prod *= lamb + i
        a = -a * prod / factorial(N)
        return schemsum(recflux(N - 1, lamb), (a, center))


def Lwconstructor(order):
    """Constructor of Lax-Wendroff scheme.

    :param order: Order of Lax Wendroff scheme
    :type order: int
    :return: Function depending on the Courant number lambda and maps to the parameters of the Lax-Wendroff scheme
    :rtype: function
    """

    def LWconstructorlamb(lamb):
        fluxschem = recflux(order, lamb)
        newfluxschem = -fluxschem[0], fluxschem[1] + 1
        susfluxschem = schemscal(schemsum(fluxschem, newfluxschem), -lamb)
        return schemsum((np.array([1]), 0), susfluxschem)

    return LWconstructorlamb
